# Copyright (c) 2013-2022 khal contributors
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""this module contains some helper functions converting strings or list of
strings to date(time) or event objects"""

import datetime as dt
import logging
import re
from calendar import isleap
from collections.abc import Callable
from time import strptime
from typing import Any

import pytz

from khal.exceptions import DateTimeParseError, FatalError

from .custom_types import LocaleConfiguration, RRuleMapType

logger = logging.getLogger("khal")


def timefstr(dtime_list: list[str], timeformat: str) -> dt.datetime:
    """converts the first item of a list (a time as a string) to a datetimeobject

    where the date is today and the time is given by a string
    removes "used" elements of list.
    """
    if len(dtime_list) == 0:
        raise ValueError()
    datetime_start = dt.datetime.strptime(dtime_list[0], timeformat)
    time_start = dt.time(*datetime_start.timetuple()[3:5])
    day_start = dt.date.today()
    dtstart = dt.datetime.combine(day_start, time_start)
    dtime_list.pop(0)
    return dtstart


def datetimefstr(
    dtime_list: list[str],
    dateformat: str,
    default_day: dt.date | None = None,
    infer_year: bool = True,
    in_future: bool = True,
) -> dt.datetime:
    """converts a datetime (as one or several string elements of a list) to
    a datetimeobject, if infer_year is True, use the `default_day`'s year as
    the year of the return datetimeobject,

    removes "used" elements of list

    example: dtime_list = ['17.03.', 'description']
             dateformat = '%d.%m.'
    or     : dtime_list = ['17.03.', '16:00', 'description']
             dateformat = '%d.%m. %H:%M'
    """
    # if now() is called as default param, mocking with freezegun won't work
    now = dt.datetime.now()
    if default_day is None:
        default_day = now.date()
    parts = dateformat.count(" ") + 1
    dtstring = " ".join(dtime_list[0:parts])
    # only time.strptime can parse the 29th of Feb. if no year is given
    dtstart_struct = strptime(dtstring, dateformat)
    if (
        infer_year
        and dtstart_struct.tm_mon == 2
        and dtstart_struct.tm_mday == 29
        and not isleap(default_day.year)
    ):
        raise ValueError

    for _ in range(parts):
        item = dtime_list.pop(0)
        if " " in item:
            logger.warning("detected a space in datetime specification, this can lead to errors.")
            logger.warning("Make sure not to quote your datetime specification.")

    if infer_year:
        dtstart = dt.datetime(*(default_day.timetuple()[:1] + dtstart_struct[1:5]))
        if in_future and dtstart < now:
            dtstart = dtstart.replace(year=dtstart.year + 1)
        if dtstart.date() < default_day:
            dtstart = dtstart.replace(year=default_day.year + 1)
        return dtstart
    else:
        return dt.datetime(*dtstart_struct[:5])


def weekdaypstr(dayname: str) -> int:
    """converts an (abbreviated) dayname to a number (mon=0, sun=6)

    :param dayname: name of abbreviation of the day
    :return: number of the day in a week
    """

    if dayname in ["monday", "mon"]:
        return 0
    if dayname in ["tuesday", "tue"]:
        return 1
    if dayname in ["wednesday", "wed"]:
        return 2
    if dayname in ["thursday", "thu"]:
        return 3
    if dayname in ["friday", "fri"]:
        return 4
    if dayname in ["saturday", "sat"]:
        return 5
    if dayname in ["sunday", "sun"]:
        return 6
    raise ValueError(f"invalid weekday name `{dayname}`")


def construct_daynames(date_: dt.date) -> str:
    """converts datetime.date into a string description

    either `Today`, `Tomorrow` or name of weekday.
    """
    if date_ == dt.date.today():
        return "Today"
    elif date_ == dt.date.today() + dt.timedelta(days=1):
        return "Tomorrow"
    else:
        return date_.strftime("%A")


def calc_day(dayname: str) -> dt.datetime:
    """converts a relative date's description to a datetime object

    :param dayname: relative day name (like 'today' or 'monday')
    :returns: date
    """
    today = dt.datetime.combine(dt.date.today(), dt.time.min)
    dayname = dayname.lower()
    if dayname == "today":
        return today
    if dayname == "tomorrow":
        return today + dt.timedelta(days=1)
    if dayname == "yesterday":
        return today - dt.timedelta(days=1)

    wday = weekdaypstr(dayname)
    days = (wday - today.weekday()) % 7
    days = 7 if days == 0 else days
    day = today + dt.timedelta(days=days)
    return day


def datefstr_weekday(dtime_list: list[str], timeformat: str, infer_year: bool) -> dt.datetime:
    """interprets first element of a list as a relative date and removes that
    element

    :param dtime_list: event description in list form
    :param timeformat: only here for compat reasons (having the same function signature)

    :returns: date
    """
    if len(dtime_list) == 0:
        raise ValueError()
    day = calc_day(dtime_list[0])
    dtime_list.pop(0)
    return day


def datetimefstr_weekday(dtime_list: list[str], timeformat: str, infer_year: bool) -> dt.datetime:
    """
    :param infer_year: only here for compat reasons (having the same function signature)
    """
    if len(dtime_list) == 0:
        raise ValueError()
    day = calc_day(dtime_list[0])
    this_time = timefstr(dtime_list[1:], timeformat)
    dtime_list.pop(0)
    dtime_list.pop(0)  # we need to pop twice as timefstr gets a copy
    dtime = dt.datetime.combine(day, this_time.time())
    return dtime


def guessdatetimefstr(
    dtime_list: list[str],
    locale: LocaleConfiguration,
    default_day: dt.date | None = None,
    in_future=True,
) -> tuple[dt.datetime, bool]:
    """
    :param in_future: if set, shortdate(time) events will be set in the future
    """
    # if now() is called as default param, mocking with freezegun won't work
    day = default_day or dt.datetime.now().date()
    # TODO rename in guessdatetimefstrLIST or something saner altogether

    def timefstr_day(dtime_list: list[str], timeformat: str, infer_year: bool) -> dt.datetime:
        if locale["timeformat"] == "%H:%M" and dtime_list[0] == "24:00":
            a_date = dt.datetime.combine(day, dt.time(0))
            dtime_list.pop(0)
        else:
            a_date = timefstr(dtime_list, timeformat)
            a_date = dt.datetime(*(day.timetuple()[:3] + a_date.timetuple()[3:5]))
        return a_date

    def datetimefwords(dtime_list: list[str], _: str, infer_year: bool) -> dt.datetime:
        if len(dtime_list) > 0 and dtime_list[0].lower() == "now":
            dtime_list.pop(0)
            return dt.datetime.now()
        raise ValueError

    def datefstr_year(dtime_list: list[str], dtformat: str, infer_year: bool) -> dt.datetime:
        return datetimefstr(dtime_list, dtformat, day, infer_year, in_future)

    dtstart = None
    fun: Callable[[list[str], str, bool], dt.datetime]
    dtformat: str
    all_day: bool
    infer_year: bool
    for fun, dtformat, all_day, infer_year in [
        (datefstr_year, locale["datetimeformat"], False, True),
        (datefstr_year, locale["longdatetimeformat"], False, False),
        (timefstr_day, locale["timeformat"], False, False),
        (datetimefstr_weekday, locale["timeformat"], False, False),
        (datefstr_year, locale["dateformat"], True, True),
        (datefstr_year, locale["longdateformat"], True, False),
        (datefstr_weekday, "", True, False),
        (datetimefwords, "", False, False),
    ]:
        # if a `short` format contains a year, treat it as a `long` format
        infer_year = infer_year and "97" not in dt.datetime(1997, 10, 11).strftime(dtformat)
        try:
            dtstart = fun(dtime_list, dtformat, infer_year=infer_year)
        except (ValueError, DateTimeParseError):
            pass
        else:
            return dtstart, all_day
    raise DateTimeParseError(
        f'Could not parse "{dtime_list}".\nPlease check your configuration '
        "or run `khal printformats` to see if this does match your configured "
        "[long](date|time|datetime)format.\nIf you suspect a bug, please "
        "file an issue at https://github.com/pimutils/khal/issues/ "
    )


def timedelta2str(delta: dt.timedelta) -> str:
    # we deliberately ignore any subsecond deltas
    total_seconds = int(abs(delta).total_seconds())

    seconds = total_seconds % 60
    total_seconds -= seconds
    total_minutes = total_seconds // 60
    minutes = total_minutes % 60
    total_minutes -= minutes
    total_hours = total_minutes // 60
    hours = total_hours % 24
    total_hours -= hours
    days = total_hours // 24

    s = []
    if days:
        s.append(str(days) + "d")
    if hours:
        s.append(str(hours) + "h")
    if minutes:
        s.append(str(minutes) + "m")
    if seconds:
        s.append(str(seconds) + "s")

    if delta != abs(delta):
        s = ["-" + part for part in s]

    return " ".join(s)


def guesstimedeltafstr(delta_string: str) -> dt.timedelta:
    """parses a timedelta from a string

    :param delta_string: string encoding time-delta, e.g. '1h 15m'
    """

    tups = re.split(r"(-?\+?\d+)", delta_string)
    if not re.match(r"^\s*$", tups[0]):
        raise ValueError(f'Invalid beginning of timedelta string "{delta_string}": "{tups[0]}"')
    tups = tups[1:]
    res = dt.timedelta()

    for num, unit in zip(tups[0::2], tups[1::2]):
        try:
            if num[0] == "+":
                num = num[1:]
            numint = int(num)
        except ValueError:
            raise DateTimeParseError(
                f'Invalid number in timedelta string "{delta_string}": "{num}"'
            )

        ulower = unit.lower().strip()
        if ulower == "d" or ulower == "day" or ulower == "days":
            res += dt.timedelta(days=numint)
        elif ulower == "h" or ulower == "hour" or ulower == "hours":
            res += dt.timedelta(hours=numint)
        elif ulower == "m" or ulower == "minute" or ulower == "minutes" or ulower == "min":
            res += dt.timedelta(minutes=numint)
        elif ulower == "s" or ulower == "second" or ulower == "seconds" or ulower == "sec":
            res += dt.timedelta(seconds=numint)
        else:
            raise ValueError(f'Invalid unit in timedelta string "{delta_string}": "{unit}"')

    return res


def guessrangefstr(
    daterange: str | list[str],
    locale: LocaleConfiguration,
    default_timedelta_date: dt.timedelta = dt.timedelta(days=1),
    default_timedelta_datetime: dt.timedelta = dt.timedelta(hours=1),
    adjust_reasonably: bool = False,
) -> tuple[dt.datetime, dt.datetime, bool]:
    """parses a range string

    :param daterange: date1 [date2 | timedelta]
    :param locale:
    :returns: start and end of the date(time) range  and if
        this is an all-day time range or not,
        **NOTE**: the end is *exclusive* if this is an allday event
    """
    range_list = daterange
    if isinstance(daterange, str):
        range_list = daterange.split(" ")
        assert isinstance(range_list, list)

    if range_list == ["week"]:
        today_weekday = dt.datetime.today().weekday()
        startdt = dt.datetime.today() - dt.timedelta(days=(today_weekday - locale["firstweekday"]))
        enddt = startdt + dt.timedelta(days=8)
        return startdt, enddt, True

    for i in reversed(range(1, len(range_list) + 1)):
        startstr = " ".join(range_list[:i])
        endstr = " ".join(range_list[i:])
        allday = False
        try:
            # figuring out start
            split = startstr.split(" ")
            start, allday = guessdatetimefstr(split, locale)
            if len(split) != 0:
                continue

            # and end
            if len(endstr) == 0:
                if allday:
                    end = start + default_timedelta_date
                else:
                    end = start + default_timedelta_datetime
            elif endstr.lower() == "eod":
                end = dt.datetime.combine(start.date(), dt.time.max)
            elif endstr.lower() == "week":
                start -= dt.timedelta(days=(start.weekday() - locale["firstweekday"]))
                end = start + dt.timedelta(days=8)
            else:
                try:
                    delta = guesstimedeltafstr(endstr)
                    if allday and delta.total_seconds() % (3600 * 24):
                        # TODO better error class, no logging in here
                        logger.fatal(
                            "Cannot give delta containing anything but whole days for allday events"
                        )
                        raise FatalError()
                    elif delta.total_seconds() == 0:
                        logger.fatal("Events that last no time are not allowed")
                        raise FatalError()

                    end = start + delta
                except (ValueError, DateTimeParseError):
                    split = endstr.split(" ")
                    end, _end_allday = guessdatetimefstr(
                        split, locale, default_day=start.date(), in_future=False
                    )
                    if len(split) != 0:
                        continue
                    if allday:
                        end += dt.timedelta(days=1)

            if adjust_reasonably:
                if allday:
                    # test if end's year is this year, but start's year is not
                    today = dt.datetime.today()
                    if end.year == today.year and start.year != today.year:
                        end = dt.datetime(start.year, *end.timetuple()[1:6])

                    if end < start:
                        end = dt.datetime(end.year + 1, *end.timetuple()[1:6])

                if end < start:
                    end = dt.datetime(*start.timetuple()[0:3] + end.timetuple()[3:5])
                if end < start:
                    end = end + dt.timedelta(days=1)
            return start, end, allday
        except (ValueError, DateTimeParseError):
            pass

    raise DateTimeParseError(
        f'Could not parse "{daterange}".\nPlease check your configuration or '
        "run `khal printformats` to see if this does match your configured "
        "[long](date|time|datetime)format.\nIf you suspect a bug, please "
        "file an issue at https://github.com/pimutils/khal/issues/ "
    )


def rrulefstr(
    repeat: str,
    until: str,
    locale: LocaleConfiguration,
    timezone: dt.tzinfo | None,
) -> RRuleMapType:
    if repeat in ["daily", "weekly", "monthly", "yearly"]:
        rrule_settings: RRuleMapType = {"freq": repeat}
        if until:
            until_dt, _ = guessdatetimefstr(until.split(" "), locale)
            if timezone:
                rrule_settings["until"] = until_dt.replace(tzinfo=timezone).astimezone(pytz.UTC)
            else:
                rrule_settings["until"] = until_dt
        return rrule_settings
    else:
        logger.fatal(
            "Invalid value for the repeat option. \
                Possible values are: daily, weekly, monthly or yearly"
        )
        raise FatalError()


def eventinfofstr(
    info_string: str,
    locale: LocaleConfiguration,
    default_event_duration: dt.timedelta,
    default_dayevent_duration: dt.timedelta,
    adjust_reasonably: bool = False,
) -> dict[str, Any]:
    """parses a string of the form START [END | DELTA] [TIMEZONE] [SUMMARY] [::
    DESCRIPTION] into a dictionary with keys: dtstart, dtend, timezone, allday,
    summary, description
    """
    description = None
    if " :: " in info_string:
        info_string, description = info_string.split(" :: ")

    parts = info_string.split(" ")
    summary = None
    start: dt.datetime | dt.date | None = None
    end: dt.datetime | dt.date | None = None
    tz: pytz.BaseTzInfo | None = None
    allday: bool = False
    for i in reversed(range(1, len(parts) + 1)):
        try:
            start, end, allday = guessrangefstr(
                " ".join(parts[0:i]),
                locale,
                default_timedelta_datetime=default_event_duration,
                default_timedelta_date=default_dayevent_duration,
                adjust_reasonably=adjust_reasonably,
            )
        except (ValueError, DateTimeParseError):
            continue
        if start is not None and end is not None:
            try:
                # next element is a valid Olson db timezone string
                tz = pytz.timezone(parts[i])
                i += 1
            except (pytz.UnknownTimeZoneError, UnicodeDecodeError, IndexError):
                tz = None
            summary = " ".join(parts[i:])
            break

    if start is None or end is None:
        raise DateTimeParseError(
            f'Could not parse "{info_string}".\nPlease check your '
            "configuration or run `khal printformats` to see if this does "
            "match your configured [long](date|time|datetime)format.\nIf you "
            "suspect a bug, please file an issue at "
            "https://github.com/pimutils/khal/issues/ "
        )

    if tz is None:
        tz = locale["default_timezone"]

    if allday:
        assert isinstance(start, dt.datetime)
        assert isinstance(end, dt.datetime)
        start = start.date()
        end = end.date()

    info: dict[str, Any] = {}
    info["dtstart"] = start
    info["dtend"] = end
    info["summary"] = summary if summary else None
    info["description"] = description
    info["timezone"] = tz if not allday else None
    info["allday"] = allday
    return info



# Copyright (c) 2013-2022 khal contributors
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

try:
    from khal.version import version
except ImportError:
    version = "invalid"
    import sys

    sys.exit(
        "Failed to find (autogenerated) version.py. This might be due to "
        "using GitHub's tarballs or svn access. Either clone "
        "from GitHub via git or get a tarball from PyPI."
    )

__productname__ = "khal"
__version__ = version
__author__ = "Christian Geier"
__copyright__ = "Copyright (c) 2013-2022 khal contributors"
__author_email__ = "khal@lostpackets.de"
__description__ = "A standards based terminal calendar"
__license__ = "Expat/MIT, see COPYING"
__homepage__ = "https://lostpackets.de/khal/"



# Copyright (c) 2013-2022 khal contributors
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""A collection of (reusable) urwid widgets

Widgets that are specific to calendaring/khal should go into __init__.py or,
if they are large, into their own files
"""

import datetime as dt
import re

import urwid


class DateConversionError(Exception):
    pass


def delete_last_word(text, number=1):
    """delete last `number` of words from text"""
    words = re.findall(r"[\w]+|[^\w\s]", text, re.UNICODE)
    for one in range(1, number + 1):
        text = text.rstrip()
        if text == "":
            return text
        text = text[: len(text) - len(words[-one])]
    return text


def delete_till_beginning_of_line(text):
    """delete till beginning of line"""
    if text.rfind("\n") == -1:
        return ""
    return text[0 : text.rfind("\n") + 1]


def delete_till_end_of_line(text):
    """delete till beginning of line"""
    if text.find("\n") == -1:
        return ""
    return text[text.find("\n") :]


def goto_beginning_of_line(text):
    if text.rfind("\n") == -1:
        return 0
    return text.rfind("\n") + 1


def goto_end_of_line(text):
    if text.find("\n") == -1:
        return len(text)
    return text.find("\n")


class ExtendedEdit(urwid.Edit):
    """A text editing widget supporting some more editing commands"""

    def keypress(self, size: tuple[int], key: str | None) -> str | None:
        if key == "ctrl w":
            self._delete_word()
        elif key == "ctrl u":
            self._delete_till_beginning_of_line()
        elif key == "ctrl k":
            self._delete_till_end_of_line()
        elif key == "ctrl a":
            self._goto_beginning_of_line()
        elif key == "ctrl e":
            self._goto_end_of_line()
        else:
            return super().keypress(size, key)

        return None

    def _delete_word(self):
        """delete word before cursor"""
        text = self.get_edit_text()
        f_text = delete_last_word(text[: self.edit_pos])
        self.set_edit_text(f_text + text[self.edit_pos :])
        self.set_edit_pos(len(f_text))

    def _delete_till_beginning_of_line(self):
        """delete till start of line before cursor"""
        text = self.get_edit_text()
        f_text = delete_till_beginning_of_line(text[: self.edit_pos])
        self.set_edit_text(f_text + text[self.edit_pos :])
        self.set_edit_pos(len(f_text))

    def _delete_till_end_of_line(self):
        """delete till end of line before cursor"""
        text = self.get_edit_text()
        f_text = delete_till_end_of_line(text[self.edit_pos :])
        self.set_edit_text(text[: self.edit_pos] + f_text)

    def _goto_beginning_of_line(self):
        text = self.get_edit_text()
        self.set_edit_pos(goto_beginning_of_line(text[: self.edit_pos]))

    def _goto_end_of_line(self):
        text = self.get_edit_text()
        self.set_edit_pos(goto_end_of_line(text[self.edit_pos :]) + self.edit_pos)


class DateTimeWidget(ExtendedEdit):
    def __init__(self, dateformat: str, on_date_change=lambda x: None, **kwargs) -> None:
        self.dateformat = dateformat
        self.on_date_change = on_date_change
        super().__init__(wrap="any", **kwargs)

    def keypress(self, size, key):
        if key == "ctrl x":
            self.decrease()
            return None
        elif key == "ctrl a":
            self.increase()
            return None

        if (
            key in ["up", "down", "tab", "shift tab", "page up", "page down", "meta enter"]
            or (key in ["right"] and self.edit_pos >= len(self.edit_text))
            or (key in ["left"] and self.edit_pos == 0)
        ):
            # when leaving the current Widget we check if currently
            # entered value is valid and if so pass the new value
            try:
                new_date = self._get_current_value()
            except DateConversionError:
                pass
            else:
                self.on_date_change(new_date)
        return super().keypress(size, key)

    def increase(self):
        """call to increase the datefield by self.timedelta"""
        self._crease(self.dtype.__add__)

    def decrease(self):
        """call to decrease the datefield by self.timedelta"""
        self._crease(self.dtype.__sub__)

    def _crease(self, fun):
        """common implementation for `self.increase` and `self.decrease`"""
        try:
            new_date = fun(self._get_current_value(), self.timedelta)
            self.on_date_change(new_date)
            self.set_edit_text(new_date.strftime(self.dateformat))
        except DateConversionError:
            pass

    def set_value(self, new_date: dt.date):
        """set a new value for this widget"""
        self.set_edit_text(new_date.strftime(self.dateformat))


class DateWidget(DateTimeWidget):
    dtype = dt.date
    timedelta = dt.timedelta(days=1)

    def _get_current_value(self):
        try:
            new_date = dt.datetime.strptime(self.get_edit_text(), self.dateformat).date()
        except ValueError:
            raise DateConversionError
        else:
            return new_date


class TimeWidget(DateTimeWidget):
    dtype = dt.datetime
    timedelta = dt.timedelta(minutes=15)

    def _get_current_value(self):
        try:
            new_datetime = dt.datetime.strptime(self.get_edit_text(), self.dateformat)
        except ValueError:
            raise DateConversionError
        else:
            return new_datetime


class Choice(urwid.PopUpLauncher):
    def __init__(
        self,
        choices: list[str],
        active: str,
        decorate_func=None,
        overlay_width: int = 32,
        callback=lambda: None,
    ) -> None:
        self.choices = choices
        self._callback = callback
        self._decorate = decorate_func or (lambda x: x)
        self._overlay_width = overlay_width
        self.active = self._original = active

    def create_pop_up(self):
        pop_up = ChoiceList(self, callback=self._callback)
        urwid.connect_signal(pop_up, "close", lambda button: self.close_pop_up())
        return pop_up

    def get_pop_up_parameters(self):
        return {
            "left": 0,
            "top": 1,
            "overlay_width": self._overlay_width,
            "overlay_height": len(self.choices),
        }

    @property
    def changed(self):
        return self._active != self._original

    @property
    def active(self):
        return self._active

    @active.setter
    def active(self, val):
        self._active = val
        self.button = urwid.Button(self._decorate(self._active))
        urwid.PopUpLauncher.__init__(self, self.button)
        urwid.connect_signal(self.button, "click", lambda button: self.open_pop_up())


class ChoiceList(urwid.WidgetWrap):
    """A pile of Button() widgets, intended to be used with Choice()"""

    signals = ["close"]

    def __init__(self, parent, callback=lambda: None) -> None:
        self.parent = parent
        self._callback = callback
        buttons = []
        for c in parent.choices:
            buttons.append(
                button(
                    parent._decorate(c),
                    attr_map="popupbg",
                    focus_map="popupbg focus",
                    on_press=self.set_choice,
                    user_data=c,
                )
            )

        pile = NPile(buttons, outermost=True)
        num = [num for num, elem in enumerate(parent.choices) if elem == parent.active][0]
        pile.focus_position = num
        fill = urwid.Filler(pile)
        urwid.WidgetWrap.__init__(self, urwid.AttrMap(fill, "popupbg"))

    def set_choice(self, button, account):
        self.parent.active = account
        self._callback()
        self._emit("close")


class SupportsNext:
    """classes inheriting from SupportsNext must implement the following methods:
    _select_first_selectable
    _select_last_selectable
    """

    def __init__(self, *args, **kwargs) -> None:
        self.outermost = kwargs.get("outermost", False)
        if "outermost" in kwargs:
            kwargs.pop("outermost")
        super().__init__(*args, **kwargs)


class NextMixin(SupportsNext):
    """Implements SupportsNext for urwid.Pile and urwid.Columns"""

    def _select_first_selectable(self):
        """select our first selectable item (recursivly if that item SupportsNext)"""
        i = self._first_selectable()
        self.focus_position = i
        if isinstance(self.contents[i][0], SupportsNext):
            self.contents[i][0]._select_first_selectable()

    def _select_last_selectable(self):
        """select our last selectable item (recursivly if that item SupportsNext)"""
        i = self._last_selectable()
        self.focus_position = i
        if isinstance(self._contents[i][0], SupportsNext):
            self.contents[i][0]._select_last_selectable()

    def _first_selectable(self):
        """return sequence number of self.contents last selectable item"""
        for j in range(len(self._contents)):
            if self._contents[j][0].selectable():
                return j
        return False

    def _last_selectable(self):
        """return sequence number of self._contents last selectable item"""
        for j in range(len(self._contents) - 1, -1, -1):
            if self._contents[j][0].selectable():
                return j
        return False

    def keypress(self, size, key):
        key = super().keypress(size, key)

        if key == "tab":
            if self.outermost and self.focus_position == self._last_selectable():
                self._select_first_selectable()
            else:
                for i in range(self.focus_position + 1, len(self._contents)):
                    if self._contents[i][0].selectable():
                        self.focus_position = i
                        if isinstance(self._contents[i][0], SupportsNext):
                            self._contents[i][0]._select_first_selectable()
                        break
                else:  # no break
                    return key
        elif key == "shift tab":
            if self.outermost and self.focus_position == self._first_selectable():
                self._select_last_selectable()
            else:
                for i in range(self.focus_position - 1, 0 - 1, -1):
                    if self._contents[i][0].selectable():
                        self.focus_position = i
                        if isinstance(self._contents[i][0], SupportsNext):
                            self._contents[i][0]._select_last_selectable()
                        break
                else:  # no break
                    return key
        else:
            return key


class NPile(NextMixin, urwid.Pile):
    pass


class NColumns(NextMixin, urwid.Columns):
    pass


class NListBox(SupportsNext, urwid.ListBox):
    def _select_first_selectable(self):
        """select our first selectable item (recursivly if that item SupportsNext)"""
        i = self._first_selectable()
        self.focus_position = i
        if isinstance(self.body[i], SupportsNext):
            self.body[i]._select_first_selectable()

    def _select_last_selectable(self):
        """select our last selectable item (recursivly if that item SupportsNext)"""
        i = self._last_selectable()
        self.focus_position = i
        if isinstance(self.body[i], SupportsNext):
            self.body[i]._select_last_selectable()

    def _first_selectable(self):
        """return sequence number of self._contents last selectable item"""
        for j in range(len(self.body)):
            if self.body[j].selectable():
                return j
        return False

    def _last_selectable(self):
        """return sequence number of self.contents last selectable item"""
        for j in range(len(self.body) - 1, -1, -1):
            if self.body[j].selectable():
                return j
        return False

    def keypress(self, size, key):
        key = super().keypress(size, key)
        if key == "tab":
            if self.outermost and self.focus_position == self._last_selectable():
                self._select_first_selectable()
            else:
                self._keypress_down(size)
        elif key == "shift tab":
            if self.outermost and self.focus_position == self._first_selectable():
                self._select_last_selectable()
            else:
                self._keypress_up(size)
        else:
            return key


class ValidatedEdit(urwid.WidgetWrap):
    def __init__(self, *args, EditWidget=ExtendedEdit, validate=False, **kwargs) -> None:
        assert validate
        self._validate_func = validate
        self._original_widget = urwid.AttrMap(EditWidget(*args, **kwargs), "edit", "edit focused")
        super().__init__(self._original_widget)

    @property
    def _get_base_widget(self):
        return self._original_widget

    @property
    def base_widget(self):
        return self._original_widget.original_widget

    def _validate(self):
        text = self.base_widget.get_edit_text()
        if self._validate_func(text):
            self._original_widget.set_attr_map({None: "edit"})
            self._original_widget.set_focus_map({None: "edit"})
            return True
        else:
            self._original_widget.set_attr_map({None: "alert"})
            self._original_widget.set_focus_map({None: "alert"})
            return False

    def get_edit_text(self):
        self._validate()
        return self.base_widget.get_edit_text()

    @property
    def edit_pos(self):
        return self.base_widget.edit_pos

    @property
    def edit_text(self):
        return self.base_widget.edit_text

    def keypress(self, size, key):
        if (
            key in ["up", "down", "tab", "shift tab", "page up", "page down", "meta enter"]
            or (key in ["right"] and self.edit_pos >= len(self.edit_text))
            or (key in ["left"] and self.edit_pos == 0)
        ):
            if not self._validate():
                return
        return super().keypress(size, key)


class PositiveIntEdit(ValidatedEdit):
    def __init__(self, *args, EditWidget=ExtendedEdit, validate=False, **kwargs) -> None:
        """Variant of Validated Edit that only accepts positive integers."""
        super().__init__(*args, validate=self._unsigned_int, **kwargs)

    @staticmethod
    def _unsigned_int(number):
        """test if `number` can be converted to a positive int"""
        try:
            return int(number) >= 0
        except ValueError:
            return False


class DurationWidget(urwid.WidgetWrap):
    @staticmethod
    def unsigned_int(number):
        """test if `number` can be converted to a positive int"""
        try:
            return int(number) >= 0
        except ValueError:
            return False

    @staticmethod
    def _convert_timedelta(dt):
        seconds = dt.total_seconds()
        days = int(seconds // (24 * 60 * 60))
        hours = int((seconds // 3600) % 24)
        minutes = int((seconds % 3600) // 60)
        seconds = int(seconds % 60)
        return days, hours, minutes, seconds

    def __init__(self, timedelta: dt.timedelta) -> None:
        days, hours, minutes, seconds = self._convert_timedelta(timedelta)

        self.days_edit = ValidatedEdit(
            edit_text=str(days), validate=self.unsigned_int, align="right"
        )
        self.hours_edit = ValidatedEdit(
            edit_text=str(hours), validate=self.unsigned_int, align="right"
        )
        self.minutes_edit = ValidatedEdit(
            edit_text=str(minutes), validate=self.unsigned_int, align="right"
        )
        self.seconds_edit = ValidatedEdit(
            edit_text=str(seconds), validate=self.unsigned_int, align="right"
        )

        self.columns = NColumns(
            [
                (4, self.days_edit),
                (2, urwid.Text("D")),
                (3, self.hours_edit),
                (2, urwid.Text("H")),
                (3, self.minutes_edit),
                (2, urwid.Text("M")),
                (3, self.seconds_edit),
                (2, urwid.Text("S")),
            ]
        )

        urwid.WidgetWrap.__init__(self, self.columns)

    def get_timedelta(self) -> dt.timedelta:
        return dt.timedelta(
            seconds=int(self.seconds_edit.get_edit_text())
            + int(self.minutes_edit.get_edit_text()) * 60
            + int(self.hours_edit.get_edit_text()) * 60 * 60
            + int(self.days_edit.get_edit_text()) * 24 * 60 * 60
        )


class AlarmsEditor(urwid.WidgetWrap):
    class AlarmEditor(urwid.WidgetWrap):
        def __init__(self, alarm: tuple[dt.timedelta, str], delete_handler) -> None:
            duration, description = alarm
            if duration.total_seconds() > 0:
                direction = "after"
            else:
                direction = "before"
                duration = -1 * duration

            self.duration = DurationWidget(duration)
            self.description = ExtendedEdit(
                edit_text=description if description is not None else ""
            )
            self.direction = Choice(["before", "after"], active=direction, overlay_width=10)
            self.columns = NColumns(
                [
                    (2, urwid.Text("  ")),
                    (21, self.duration),
                    (14, urwid.Padding(self.direction, right=1)),
                    self.description,
                    (10, button("Delete", on_press=delete_handler, user_data=self)),
                ]
            )

            urwid.WidgetWrap.__init__(self, self.columns)

        def get_alarm(self):
            direction = self.direction.active
            if direction == "before":
                prefix = -1
            else:
                prefix = 1
            return (prefix * self.duration.get_timedelta(), self.description.get_edit_text())

    def __init__(self, event) -> None:
        self.event = event

        self.pile = NPile(
            [urwid.Text("Alarms:")]
            + [self.AlarmEditor(a, self.remove_alarm) for a in event.alarms]
            + [urwid.Columns([(12, button("Add", on_press=self.add_alarm))])]
        )

        urwid.WidgetWrap.__init__(self, self.pile)

    def clear(self) -> None:
        """clear the alarm list"""
        self.pile.contents.clear()

    def add_alarm(self, button, timedelta: dt.timedelta | None = None):
        if timedelta is None:
            timedelta = dt.timedelta(0)
        self.pile.contents.insert(
            len(self.pile.contents) - 1,
            (self.AlarmEditor((timedelta, self.event.summary), self.remove_alarm), ("weight", 1)),
        )

    def remove_alarm(self, button, editor):
        self.pile.contents.remove((editor, ("weight", 1)))

    def get_alarms(self):
        alarms = []
        for widget, _ in self.pile.contents:
            if isinstance(widget, self.AlarmEditor):
                alarms.append(widget.get_alarm())
        return alarms

    @property
    def changed(self):
        try:
            return self.event.alarms != self.get_alarms()
        except ValueError:
            return False


class FocusLineBoxWidth(urwid.WidgetDecoration, urwid.WidgetWrap):
    def __init__(self, widget) -> None:
        # we cheat here with the attrs, if we use thick dividers we apply the
        # focus attr group. We probably should fix this in render()
        hline = urwid.AttrMap(urwid.Divider("─"), "frame")
        hline_focus = urwid.AttrMap(urwid.Divider("━"), "frame focus")
        self._vline = urwid.AttrMap(urwid.SolidFill("│"), "frame")
        self._vline_focus = urwid.AttrMap(urwid.SolidFill("┃"), "frame focus")
        self._topline = urwid.Columns(
            [
                ("fixed", 1, urwid.AttrMap(urwid.Text("┌"), "frame")),
                hline,
                ("fixed", 1, urwid.AttrMap(urwid.Text("┐"), "frame")),
            ]
        )
        self._topline_focus = urwid.Columns(
            [
                ("fixed", 1, urwid.AttrMap(urwid.Text("┏"), "frame focus")),
                hline_focus,
                ("fixed", 1, urwid.AttrMap(urwid.Text("┓"), "frame focus")),
            ]
        )
        self._bottomline = urwid.Columns(
            [
                ("fixed", 1, urwid.AttrMap(urwid.Text("└"), "frame")),
                hline,
                ("fixed", 1, urwid.AttrMap(urwid.Text("┘"), "frame")),
            ]
        )
        self._bottomline_focus = urwid.Columns(
            [
                ("fixed", 1, urwid.AttrMap(urwid.Text("┗"), "frame focus")),
                hline_focus,
                ("fixed", 1, urwid.AttrMap(urwid.Text("┛"), "frame focus")),
            ]
        )
        self._middle = urwid.Columns(
            [("fixed", 1, self._vline), widget, ("fixed", 1, self._vline)],
            focus_column=1,
        )
        self._all = urwid.Pile(
            [("flow", self._topline), self._middle, ("flow", self._bottomline)],
            focus_item=1,
        )

        urwid.WidgetWrap.__init__(self, self._all)
        self._original_widget = widget

    def render(self, size, focus):
        inner = self._all.contents[1][0]
        if focus:
            self._all.contents[0] = (self._topline_focus, ("pack", None))
            inner.contents[0] = (self._vline_focus, ("given", 1, False))
            inner.contents[2] = (self._vline_focus, ("given", 1, False))
            self._all.contents[2] = (self._bottomline_focus, ("pack", None))
        else:
            self._all.contents[0] = (self._topline, ("pack", None))
            inner.contents[0] = (self._vline, ("given", 1, False))
            inner.contents[2] = (self._vline, ("given", 1, False))
            self._all.contents[2] = (self._bottomline, ("pack", None))
        return super().render(size, focus)


class FocusLineBoxColor(urwid.WidgetDecoration, urwid.WidgetWrap):
    def __init__(self, widget) -> None:
        hline = urwid.Divider("─")
        self._vline = urwid.AttrMap(urwid.SolidFill("│"), "frame")
        self._topline = urwid.AttrMap(
            urwid.Columns(
                [
                    ("fixed", 1, urwid.Text("┌")),
                    hline,
                    ("fixed", 1, urwid.Text("┐")),
                ]
            ),
            "frame",
        )
        self._bottomline = urwid.AttrMap(
            urwid.Columns(
                [
                    ("fixed", 1, urwid.Text("└")),
                    hline,
                    ("fixed", 1, urwid.Text("┘")),
                ]
            ),
            "frame",
        )

        self._middle = urwid.Columns(
            [("fixed", 1, self._vline), widget, ("fixed", 1, self._vline)],
            focus_column=1,
        )
        self._all = urwid.Pile(
            [("flow", self._topline), self._middle, ("flow", self._bottomline)],
            focus_item=1,
        )

        urwid.WidgetWrap.__init__(self, self._all)
        self._original_widget = widget

    def render(self, size, focus):
        if focus:
            self._middle.contents[0][0].set_attr_map({None: "frame focus color"})
            self._all.contents[0][0].set_attr_map({None: "frame focus color"})
            self._all.contents[2][0].set_attr_map({None: "frame focus color"})
        else:
            self._middle.contents[0][0].set_attr_map({None: "frame"})
            self._all.contents[0][0].set_attr_map({None: "frame"})
            self._all.contents[2][0].set_attr_map({None: "frame"})
        return super().render(size, focus)


class FocusLineBoxTop(urwid.WidgetDecoration, urwid.WidgetWrap):
    def __init__(self, widget) -> None:
        topline = urwid.AttrMap(urwid.Divider("━"), "frame")
        self._all = urwid.Pile([("flow", topline), widget], focus_item=1)
        urwid.WidgetWrap.__init__(self, self._all)
        self._original_widget = widget

    def render(self, size, focus):
        if focus:
            self._all.contents[0][0].set_attr_map({None: "frame focus top"})
        else:
            self._all.contents[0][0].set_attr_map({None: "frame"})
        return super().render(size, focus)


linebox = {
    "color": FocusLineBoxColor,
    "top": FocusLineBoxTop,
    "width": FocusLineBoxWidth,
    "False": urwid.WidgetPlaceholder,
}


def button(
    *args,
    attr_map: str = "button",
    focus_map="button focus",
    padding_left=0,
    padding_right=0,
    **kwargs,
):
    """wrapping an urwid button in attrmap and padding"""
    button_ = urwid.Button(*args, **kwargs)
    button_ = urwid.AttrMap(button_, attr_map=attr_map, focus_map=focus_map)
    button_ = urwid.Padding(button_, left=padding_left, right=padding_right)
    return button_


class CAttrMap(urwid.AttrMap):
    """A variant of AttrMap that exposes all properties of the original widget"""

    def __getattr__(self, name):
        return getattr(self.original_widget, name)


class CPadding(urwid.Padding):
    """A variant of Patting that exposes some properties of the original widget"""

    @property
    def active(self):
        return self.original_widget.active

    @property
    def changed(self):
        return self.original_widget.changed



# Copyright (c) 2013-2022 khal contributors
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import datetime as dt
import logging
import signal
import sys
from enum import IntEnum
from typing import Literal

import click
import urwid

from khal import plugins, utils
from khal.khalendar import CalendarCollection
from khal.khalendar.exceptions import FatalError, ReadOnlyCalendarError
from khal.parse_datetime import timedelta2str

from . import colors
from .base import Pane, Window
from .calendarwidget import CalendarWidget
from .editor import EventEditor, ExportDialog
from .widgets import CAttrMap, NColumns, NPile, button, linebox
from .widgets import ExtendedEdit as Edit

logger = logging.getLogger("khal")

#  Overview of how this all meant to fit together:
#
#   ┌──ClassicView(Pane)────────────────────────────────────────┐
#   │                                                           │
#   │ ┌─CalendarWidget──┐  ┌────EventColumn───────────────────┐ │
#   │ │                 │  │                                  │ │
#   │ │                 │  │ ┌─DListBox─────────────────────┐ │ │
#   │ │                 │  │ │                              │ │ │
#   │ │                 │  │ │ ┌─DayWalker────────────────┐ │ │ │
#   │ │                 │  │ │ │                          │ │ │ │
#   │ │                 │  │ │ │ ┌─BoxAdapter───────────┐ │ │ │ │
#   │ │                 │  │ │ │ │                      │ │ │ │ │
#   │ │                 │  │ │ │ │ ┌─DateListBox──────┐ │ │ │ │ │
#   │ │                 │  │ │ │ │ │ DateHeader       │ │ │ │ │ │
#   │ │                 │  │ │ │ │ │ U_Event          │ │ │ │ │ │
#   │ │                 │  │ │ │ │ │  ...             │ │ │ │ │ │
#   │ │                 │  │ │ │ │ │ U_Event          │ │ │ │ │ │
#   │ │                 │  │ │ │ │ └──────────────────┘ │ │ │ │ │
#   │ │                 │  │ │ │ └──────────────────────┘ │ │ │ │
#   │ │                 │  │ │ │       ...                │ │ │ │
#   │ │                 │  │ │ │ ┌─BoxAdapter───────────┐ │ │ │ │
#   │ │                 │  │ │ │ │                      │ │ │ │ │
#   │ │                 │  │ │ │ │ ┌─DateListBox──────┐ │ │ │ │ │
#   │ │                 │  │ │ │ │ │ DateHeader       │ │ │ │ │ │
#   │ │                 │  │ │ │ │ │ U_Event          │ │ │ │ │ │
#   │ │                 │  │ │ │ │ │  ...             │ │ │ │ │ │
#   │ │                 │  │ │ │ │ │ U_Event          │ │ │ │ │ │
#   │ │                 │  │ │ │ │ └──────────────────┘ │ │ │ │ │
#   │ │                 │  │ │ │ └──────────────────────┘ │ │ │ │
#   │ │                 │  │ │ └──────────────────────────┘ │ │ │
#   │ │                 │  │ └──────────────────────────────┘ │ │
#   │ └─────────────────┘  └──────────────────────────────────┘ │
#   └───────────────────────────────────────────────────────────┘


class DeletionType(IntEnum):
    ALL = 0
    INSTANCES = 1


class DateConversionError(Exception):
    pass


class SelectableText(urwid.Text):
    def selectable(self) -> bool:
        return True

    def keypress(self, size: tuple[int], key: str | None) -> str | None:
        return key

    def get_cursor_coords(self, size: tuple[int]) -> tuple[int, int]:
        return 0, 0

    def render(self, size, focus=False):
        canv = super().render(size, focus)
        if focus:
            canv = urwid.CompositeCanvas(canv)
            canv.cursor = 0, 0
        return canv


class DateHeader(SelectableText):
    def __init__(self, day: dt.date, dateformat: str, conf) -> None:
        """
        :param day: the date that is represented by this DateHeader instance
        :param dateformat: format to print `day` in
        """
        self._day = day
        self._dateformat = dateformat
        self._conf = conf
        super().__init__("", wrap="clip")
        self.update_date_line()

    def update_date_line(self) -> None:
        """update self, so that the timedelta is accurate

        to be called after a date change
        """
        self.set_text(self.relative_day(self._day, self._dateformat))

    def relative_day(self, day: dt.date, dtformat: str) -> str:
        """convert day into a string with its weekday and relative distance to today

        :param day: day to be converted
        :param dtformat: the format day is to be printed in, passed to strftime
        """

        weekday = day.strftime("%A")
        daystr = day.strftime(dtformat)
        if day == dt.date.today():
            return f"Today ({weekday}, {daystr})"
        elif day == dt.date.today() + dt.timedelta(days=1):
            return f"Tomorrow ({weekday}, {daystr})"
        elif day == dt.date.today() - dt.timedelta(days=1):
            return f"Yesterday ({weekday}, {daystr})"

        approx_delta = utils.relative_timedelta_str(day)

        return f"{weekday}, {daystr} ({approx_delta})"

    def keypress(self, size: tuple[int], key: str | None) -> str | None:
        binds = self._conf["keybindings"]
        if key in binds["left"]:
            key = "left"
        elif key in binds["up"]:
            key = "up"
        elif key in binds["right"]:
            key = "right"
        elif key in binds["down"]:
            key = "down"
        return key


class U_Event(urwid.Text):
    def __init__(self, event, conf, delete_status, this_date=None, relative=True) -> None:
        """representation of an event in EventList

        :param event: the encapsulated event
        :type event: khal.event.Event
        """
        if relative:
            if isinstance(this_date, dt.datetime) or not isinstance(this_date, dt.date):
                raise ValueError(
                    f"`this_date` is of type `{type(this_date)}`, should be `datetime.date`"
                )
        self.event = event
        self.delete_status = delete_status
        self.this_date = this_date
        self._conf = conf
        self.relative = relative
        super().__init__("", wrap="clip")
        self.set_title()

    def get_cursor_coords(self, size) -> tuple[int, int]:
        return 0, 0

    def render(self, size, focus=False):
        canv = super().render(size, focus)
        if focus:
            canv = urwid.CompositeCanvas(canv)
            canv.cursor = 0, 0
        return canv

    @classmethod
    def selectable(cls) -> bool:
        return True

    @property
    def uid(self) -> str:
        return self.event.calendar + "\n" + str(self.event.href) + "\n" + str(self.event.etag)

    @property
    def recuid(self) -> tuple[str, str]:
        return (self.uid, self.event.recurrence_id)

    def set_title(self, mark: str = " ") -> None:
        mark = {
            DeletionType.ALL: "D",
            DeletionType.INSTANCES: "d",
            None: "",
        }[self.delete_status(self.recuid)]
        if self.relative:
            format_ = self._conf["view"]["agenda_event_format"]
        else:
            format_ = self._conf["view"]["event_format"]
        formatter_ = utils.human_formatter(format_, colors=False)
        if self.this_date:
            date_ = self.this_date
        elif self.event.allday:
            date_ = self.event.start
        else:
            date_ = self.event.start.date()
        text = formatter_(self.event.attributes(date_, colors=False))
        if self._conf["locale"]["unicode_symbols"]:
            newline = " \N{LEFTWARDS ARROW WITH HOOK} "
        else:
            newline = " -- "

        self.set_text(mark + " " + text.replace("\n", newline))

    def keypress(self, size: tuple[int], key: str | None) -> str | None:
        binds = self._conf["keybindings"]
        if key in binds["left"]:
            key = "left"
        elif key in binds["up"]:
            key = "up"
        elif key in binds["right"]:
            key = "right"
        elif key in binds["down"]:
            key = "down"
        return key


class EventListBox(urwid.ListBox):
    """Container for list of U_Events"""

    def __init__(
        self,
        *args,
        parent,
        conf,
        delete_status,
        toggle_delete_instance,
        toggle_delete_all,
        set_focus_date_callback=None,
        **kwargs,
    ) -> None:
        self._init: bool = True
        self.parent: ClassicView = parent
        self.delete_status = delete_status
        self.toggle_delete_instance = toggle_delete_instance
        self.toggle_delete_all = toggle_delete_all
        self._conf = conf
        self._old_focus = None
        self.set_focus_date_callback = set_focus_date_callback
        super().__init__(*args, **kwargs)

    def keypress(self, size: tuple[int], key: str | None) -> str | None:
        return super().keypress(size, key)

    @property
    def focus_event(self) -> U_Event | None:
        if self.focus is None:
            return None
        else:
            return self.focus.original_widget

    def refresh_titles(self, min_date, max_date, everything):
        """Refresh only the currently focused event's title

        as we currently only use `EventListBox` in search and there we can only
        modify the currently focused event, no real implementation is needad at
        this time

        ignores all arguments
        """
        self.focus.original_widget.set_title()


class DListBox(EventListBox):
    """Container for a DayWalker"""

    # XXX unfortunate naming, there is also DateListBox

    def __init__(self, *args, **kwargs) -> None:
        dynamic_days = kwargs.pop("dynamic_days", True)
        super().__init__(*args, **kwargs)
        self._init = dynamic_days

    def render(self, size, focus=False):
        if self._init:
            self._init = False
        while not isinstance(self.body, StaticDayWalker) and "bottom" in self.ends_visible(size):
            self.body._autoextend()
        return super().render(size, focus)

    def clean(self):
        """reset event most recently in focus"""
        if self._old_focus is not None:
            try:
                self.body[self._old_focus].body[0].set_attr_map({None: "date"})
            except IndexError:
                # after reseting the EventList, the old focus might not exist
                pass

    def ensure_date(self, day: dt.date) -> None:
        """ensure an entry for `day` exists and bring it into focus"""
        try:
            self._old_focus = self.focus_position
        except IndexError:
            pass
        self.body.ensure_date(day)
        self.clean()

    def keypress(self, size: tuple[int], key: str | None) -> str | None:
        if key in self._conf["keybindings"]["up"]:
            key = "up"
        if key in self._conf["keybindings"]["down"]:
            key = "down"

        if key in self._conf["keybindings"]["today"]:
            self.parent.calendar.base_widget.set_focus_date(dt.date.today())

        rval = super().keypress(size, key)
        self.clean()
        if key in ["up", "down"]:
            try:
                self._old_focus = self.focus_position
            except IndexError:
                pass
            day = self.body[self.body.focus].date

            # we need to save DateListBox.selected_date and reset it later, because
            # calling CalendarWalker.set_focus_date() calls back into
            # DayWalker().update_by_date() which actually sets selected_date
            # that's why it's called callback hell...
            currently_selected_date = DateListBox.selected_date
            self.set_focus_date_callback(day)  # TODO convert to callback
            DateListBox.selected_date = currently_selected_date
        return rval

    @property
    def focus_event(self):
        return self.body.focus_event

    @property
    def current_date(self) -> dt.date:
        return self.body.current_day

    def refresh_titles(self, start, end, recurring):
        self.body.refresh_titles(start, end, recurring)

    def update_date_line(self):
        self.body.update_date_line()


class DayWalker(urwid.SimpleFocusListWalker):
    """A list Walker that contains a list of DateListBox objects, each representing
    one day and associated events"""

    def __init__(self, this_date, eventcolumn, conf, collection, delete_status) -> None:
        self.eventcolumn = eventcolumn
        self._conf = conf
        self.delete_status = delete_status
        self._init = True
        self._last_day = this_date
        self._first_day = this_date
        self._collection = collection

        super().__init__([])
        self.ensure_date(this_date)

    def reset(self):
        """delete all events contained in this DayWalker"""
        self.clear()
        self._last_day = None
        self._first_day = None

    def ensure_date(self, day: dt.date) -> None:
        """make sure a DateListBox for `day` exists, update it and bring it into focus"""
        # TODO this function gets called twice on every date change, not necessary but
        # isn't very costly either
        if self.days_to_next_already_loaded(day) > 200:  # arbitrary number
            self.reset()
        item_no = None
        if len(self) == 0:
            pile = self._get_events(day)
            self.append(pile)
            self._last_day = day
            self._first_day = day
            item_no = 0
        while day < self[0].date:
            self._autoprepend()
            item_no = 0
        while day > self[-1].date:
            self._autoextend()
            item_no = len(self) - 1
        if item_no is None:
            item_no = (day - self[0].date).days

        assert self[item_no].date == day
        self[item_no].set_selected_date(day)
        self.set_focus(item_no)

    def days_to_next_already_loaded(self, day: dt.date) -> int:
        """return number of days until `day` is already loaded into the CalendarWidget"""
        if len(self) == 0:
            return 0
        elif self[0].date <= day <= self[-1].date:
            return 0
        elif day <= self[0].date:
            return (self[0].date - day).days
        elif self[-1].date <= day:
            return (day - self[-1].date).days
        else:
            raise ValueError("This should not happen")

    def update_events_ondate(self, day):
        """refresh the contents of the day's DateListBox"""
        offset = (day - self[0].date).days
        assert self[offset].date == day
        self[offset] = self._get_events(day)

    def refresh_titles(self, start: dt.date, end: dt.date, everything: bool):
        """refresh events' titles

        if `everything` is True, reset all titles, otherwise only
        those between `start` and `end`
        """
        start = start.date() if isinstance(start, dt.datetime) else start
        end = end.date() if isinstance(end, dt.datetime) else end

        if everything:
            start = self[0].date
            end = self[-1].date
        else:
            start = max(self[0].date, start)
            end = min(self[-1].date, end)

        offset = (start - self[0].date).days
        length = (end - start).days
        for index in range(offset, offset + length + 1):
            self[index].refresh_titles()

    def update_range(self, start: dt.date, end: dt.date, everything: bool = False):
        """refresh contents of all days between start and end (inclusive)"""
        start = start.date() if isinstance(start, dt.datetime) else start
        end = end.date() if isinstance(end, dt.datetime) else end

        if everything:
            start = self[0].date
            end = self[-1].date
        else:
            start = max(self[0].date, start)
            end = min(self[-1].date, end)

        day = start
        while day <= end:
            self.update_events_ondate(day)
            day += dt.timedelta(days=1)

    def update_date_line(self):
        for one in self:
            one.update_date_line()

    def set_focus(self, position):
        """set focus by item number"""
        while position >= len(self) - 1:
            self._autoextend()
        while position <= 0:
            self._autoprepend()
            position += 1
        return super().set_focus(position)

    def _autoextend(self):
        self._last_day += dt.timedelta(days=1)
        pile = self._get_events(self._last_day)
        self.append(pile)

    def _autoprepend(self):
        """prepend the day before the first day to ourself"""
        # we need to actively reset the last element's attribute, as their
        # render() method does not get called otherwise, and they would
        # be indicated as the currently selected date
        self[self.focus or 0].reset_style()
        self._first_day -= dt.timedelta(days=1)
        pile = self._get_events(self._first_day)
        self.insert(0, pile)

    def _get_events(self, day: dt.date) -> urwid.Widget:
        """get all events on day, return a DateListBox of `U_Event()`s"""
        event_list = []
        date_header = DateHeader(
            day=day,
            dateformat=self._conf["locale"]["longdateformat"],
            conf=self._conf,
        )
        event_list.append(urwid.AttrMap(date_header, "date"))
        self.events = sorted(self._collection.get_events_on(day))
        event_list.extend(
            [
                urwid.AttrMap(
                    U_Event(
                        event, conf=self._conf, this_date=day, delete_status=self.delete_status
                    ),
                    "calendar " + event.calendar,
                    "reveal focus",
                )
                for event in self.events
            ]
        )
        return urwid.BoxAdapter(
            DateListBox(urwid.SimpleFocusListWalker(event_list), date=day),
            (len(event_list) + 1) if self.events else 1,
        )

    def selectable(self) -> bool:
        """mark this widget as selectable"""
        return True

    @property
    def focus_event(self) -> U_Event:
        return self[self.focus].original_widget.focus_event

    @property
    def current_day(self) -> dt.date:
        return self[self.focus].original_widget.date

    @property
    def first_date(self) -> dt.date:
        return self[0].original_widget.date

    @property
    def last_date(self) -> dt.date:
        return self[-1].original_widget.date


class StaticDayWalker(DayWalker):
    """Only show events for a fixed number of days."""

    def ensure_date(self, day: dt.date) -> None:
        """make sure a DateListBox for `day` exists, update it and bring it into focus"""
        # TODO cache events for each day and update as needed
        num_days = max(1, self._conf["default"]["timedelta"].days)

        for delta in range(num_days):
            pile = self._get_events(day + dt.timedelta(days=delta))
            if len(self) <= delta:
                self.append(pile)
            else:
                self[delta] = pile
        assert self[0].date == day

    def update_events_ondate(self, day):
        """refresh the contents of the day's DateListBox"""
        self[0] = self._get_events(day)

    def refresh_titles(self, start: dt.date, end: dt.date, everything: bool) -> None:
        """refresh events' titles

        if `everything` is True, reset all titles, otherwise only
        those between `start` and `end`
        """
        # TODO: why are we not using the arguments?
        for one in self:
            one.refresh_titles()

    def update_range(self, start: dt.date, end: dt.date, everything: bool = False):
        """refresh contents of all days between start and end (inclusive)"""
        start = start.date() if isinstance(start, dt.datetime) else start
        end = end.date() if isinstance(end, dt.datetime) else end

        update = everything
        for one in self:
            if start <= one.date <= end:
                update = True
        if update:
            self.ensure_date(self[0].date)

    def set_focus(self, position):
        """set focus by item number"""
        return urwid.SimpleFocusListWalker.set_focus(self, position)


class DateListBox(urwid.ListBox):
    """A ListBox container containing all events for one specific date
    used with a SimpleFocusListWalker
    """

    selected_date = None

    def __init__(self, content, date) -> None:
        self.date = date
        super().__init__(content)

    def __repr__(self) -> str:
        return f"<DateListBox {self.date}>"

    __str__ = __repr__

    def render(self, size, focus):
        if focus:
            self.body[0].set_attr_map({None: "date header focused"})
        elif DateListBox.selected_date == self.date:
            self.body[0].set_attr_map({None: "date header selected"})
        else:
            self.reset_style()
        return super().render(size, focus)

    def reset_style(self):
        self.body[0].set_attr_map({None: "date header"})

    def set_selected_date(self, day: dt.date) -> None:
        """Mark `day` as selected

        :param day: day to mark as selected
        """
        DateListBox.selected_date = day
        # we need to touch the title's content to make sure
        # that urwid re-renders the title
        title = self.body[0].original_widget
        title.set_text(title.get_text()[0])

    @property
    def focus_event(self) -> U_Event | None:
        if self.body.focus == 0:
            return None
        else:
            return self.focus.original_widget

    def refresh_titles(self):
        """refresh the titles of all events"""
        for uevent in self.body[1:]:
            if isinstance(uevent._original_widget, U_Event):
                uevent.original_widget.set_title()

    def update_date_line(self):
        """update the date text in the first line, e.g., if the current date changed"""
        self.body[0].original_widget.update_date_line()


class EventColumn(urwid.WidgetWrap):
    """Container for list of events

    Handles modifying events, showing events' details and editing them
    """

    def __init__(self, elistbox, pane) -> None:
        self.pane = pane
        self._conf = pane._conf
        self.divider = urwid.Divider("─")
        self.editor = False
        self._last_focused_date: dt.date | None = None
        self._eventshown: tuple[str, str] | None = None
        self.event_width = int(self.pane._conf["view"]["event_view_weighting"])
        self.delete_status = pane.delete_status
        self.toggle_delete_all = pane.toggle_delete_all
        self.toggle_delete_instance = pane.toggle_delete_instance
        self.dlistbox: DListBox = elistbox
        self.container = urwid.Pile([self.dlistbox])
        urwid.WidgetWrap.__init__(self, self.container)

    @property
    def focus_event(self) -> U_Event | None:
        """returns the event currently in focus"""
        return self.dlistbox.focus_event

    def view(self, event):
        """show event in the lower part of this column"""
        self.container.contents.append((self.divider, ("pack", None)))
        self.container.contents.append(
            (
                EventDisplay(self.pane._conf, event, collection=self.pane.collection),
                ("weight", self.event_width),
            )
        )

    def clear_event_view(self):
        while len(self.container.contents) > 1:
            self.container.contents.pop()

    def set_focus_date(self, date):
        """We need this, so we can use it as a callback"""
        self.focus_date = date

    @property
    def focus_date(self) -> dt.date:
        return self.dlistbox.current_date

    @focus_date.setter
    def focus_date(self, date: dt.date) -> None:
        self._last_focused_date = date
        self.dlistbox.ensure_date(date)

    def update(self, min_date, max_date: dt.date, everything: bool):
        """update DateListBox

        if `everything` is True, reset all displayed dates, else only those between
        min_date and max_date
        """
        if everything:
            min_date = self.pane.calendar.base_widget.walker.earliest_date
            max_date = self.pane.calendar.base_widget.walker.latest_date
        self.pane.base_widget.calendar.base_widget.reset_styles_range(min_date, max_date)
        if everything:
            min_date = self.dlistbox.body.first_date
            max_date = self.dlistbox.body.last_date
        self.dlistbox.body.update_range(min_date, max_date)

    def refresh_titles(self, min_date: dt.date, max_date: dt.date, everything: bool) -> None:
        """refresh titles in DateListBoxes

        if `everything` is True, reset all displayed dates, else only those between
        min_date and max_date
        """
        self.dlistbox.refresh_titles(min_date, max_date, everything)

    def update_date_line(self) -> None:
        """refresh titles in DateListBoxes"""
        self.dlistbox.update_date_line()

    def edit(self, event, always_save: bool = False, external_edit: bool = False) -> None:
        """create an EventEditor and display it

        :param event: event to edit
        :type event: khal.event.Event
        :param always_save: even save the event if it hasn't changed
        """
        if event.readonly:
            self.pane.window.alert(("alert", f"Calendar `{event.calendar}` is read-only."))
            return

        if isinstance(event.start_local, dt.datetime):
            original_start = event.start_local.date()
        else:
            original_start = event.start_local
        if isinstance(event.end_local, dt.datetime):
            original_end = event.end_local.date()
        else:
            original_end = event.end_local

        def update_colors(new_start: dt.date, new_end: dt.date, everything: bool = False):
            """reset colors in the calendar widget and dates in DayWalker
            between min(new_start, original_start)

            :param everything: set to True if event is a recurring one, than everything
                  gets reseted
            """
            # TODO cleverer support for recurring events, where more than start and
            # end dates are affected (complicated)
            if isinstance(new_start, dt.datetime):
                new_start = new_start.date()
            if isinstance(new_end, dt.datetime):
                new_end = new_end.date()
            start = min(original_start, new_start)
            end = max(original_end, new_end)
            self.pane.eventscolumn.base_widget.update(start, end, everything)

            # set original focus date
            self.pane.calendar.original_widget.set_focus_date(new_start)
            self.pane.eventscolumn.original_widget.set_focus_date(new_start)

        if self.editor:
            self.pane.window.backtrack()

        assert not self.editor
        if external_edit:
            self.pane.window.loop.stop()
            ics = click.edit(event.raw, extension=".ics")
            self.pane.window.loop.start()
            if ics is None:
                return
            # KeyErrors can occur here when we destroy DTSTART,
            # otherwise, even broken .ics files seem to be no problem
            new_event = self.pane.collection.create_event_from_ics(
                ics,
                etag=event.etag,
                calendar_name=event.calendar,
                href=event.href,
            )
            self.pane.collection.update(new_event)
            update_colors(
                new_event.start_local, new_event.end_local, (event.recurring or new_event.recurring)
            )
        else:
            self.editor = True
            editor = EventEditor(self.pane, event, update_colors, always_save=always_save)

            ContainerWidget = linebox[self.pane._conf["view"]["frame"]]
            new_pane = urwid.Columns(
                [
                    ("weight", 2, CAttrMap(ContainerWidget(editor), "editor", "editor focus")),
                    ("weight", 1, CAttrMap(ContainerWidget(self.dlistbox), "eventcolumn")),
                ],
                dividechars=0,
                focus_column=0,
            )
            new_pane.title = editor.title

            def teardown(data):
                self.editor = False

            self.pane.window.open(new_pane, callback=teardown)

    def export_event(self):
        """export the event in focus as an ICS file"""

        def export_this(_, user_data):
            try:
                self.focus_event.event.export_ics(user_data.get_edit_text())
            except Exception as error:
                self.pane.window.backtrack()
                self.pane.window.alert(("alert", f"Failed to save event: {error}"))
            else:
                self.pane.window.backtrack()
                self.pane.window.alert("Event successfully exported")

        overlay = urwid.Overlay(
            ExportDialog(
                export_this,
                self.pane.window.backtrack,
                self.focus_event.event,
            ),
            self.pane,
            "center",
            ("relative", 50),
            ("relative", 50),
            None,
        )
        self.pane.window.open(overlay)

    def toggle_delete(self):
        """toggle the delete status of the event in focus"""
        event = self.focus_event

        def delete_this(_):
            self.toggle_delete_instance(event.recuid)
            self.pane.window.backtrack()
            self.refresh_titles(
                event.event.start_local, event.event.end_local, event.event.recurring
            )

        def delete_all(_):
            self.toggle_delete_all(event.recuid)
            self.pane.window.backtrack()
            self.refresh_titles(
                event.event.start_local, event.event.end_local, event.event.recurring
            )

        if event.event.readonly:
            self.pane.window.alert(
                ("alert", f"Calendar {event.event.calendar} is read-only."),
            )
            return
        status = self.delete_status(event.recuid)
        refresh = True
        if status == DeletionType.ALL:
            self.toggle_delete_all(event.recuid)
        elif status == DeletionType.INSTANCES:
            self.toggle_delete_instance(event.recuid)
        elif event.event.recurring:
            # FIXME if in search results, original pane is used for overlay, not search results
            # also see issue of reseting titles below, probably related
            self.pane.dialog(
                text="This is a recurring event.\nWhich instances do you want to delete?",
                buttons=[
                    ("Only this", delete_this),
                    ("All (past and future)", delete_all),
                    ("Abort", self.pane.window.backtrack),
                ],
            )
            refresh = False
        else:
            self.toggle_delete_all(event.recuid)
        if refresh:
            self.refresh_titles(
                event.event.start_local, event.event.end_local, event.event.recurring
            )
            event.set_title()  # if we are in search results, refresh_titles doesn't work properly

    def duplicate(self) -> None:
        """duplicate the event in focus"""
        # TODO copying from birthday calendars is currently problematic
        # because their title is determined by X-BIRTHDAY and X-FNAME properties
        # which are also copied. If the events' summary is edited it will show
        # up on disk but not be displayed in khal
        if self.focus_event is None:
            return None
        event = self.focus_event.event.duplicate()
        try:
            self.pane.collection.insert(event)
        except ReadOnlyCalendarError:
            event.calendar = (
                self.pane.collection.default_calendar_name or self.pane.collection.writable_names[0]
            )
            self.edit(event, always_save=True)
        start_date, end_date = event.start_local, event.end_local
        if isinstance(start_date, dt.datetime):
            start_date = start_date.date()
        if isinstance(end_date, dt.datetime):
            end_date = end_date.date()
        self.pane.eventscolumn.base_widget.update(start_date, end_date, event.recurring)
        try:
            self._old_focus = self.focus_position
        except IndexError:
            pass

    def new(self, date: dt.date, end: dt.date | None = None) -> None:
        """create a new event on `date` at the next full hour and edit it

        :param date: default date for new event
        :param end: date the event ends on (inclusive)
        """
        dtstart: dt.date
        dtend: dt.date
        if not self.pane.collection.writable_names:
            self.pane.window.alert(("alert", "No writable calendar."))
            return
        if date is None:
            date = dt.datetime.now()
        if end is None:
            dtstart = dt.datetime.combine(date, dt.time(dt.datetime.now().hour))
            dtend = dtstart + dt.timedelta(minutes=60)
            allday = False
        else:  # TODO XXX why do we use allday if end is not set???
            dtstart = date
            dtend = end + dt.timedelta(days=1)
            allday = True
        event = self.pane.collection.create_event_from_dict(
            {
                "dtstart": dtstart,
                "dtend": dtend,
                "summary": "",
                "timezone": self._conf["locale"]["default_timezone"],
                "allday": allday,
                "alarms": timedelta2str(self._conf["default"]["default_event_alarm"]),
            }
        )
        self.edit(event)

    def selectable(self):
        return True

    def keypress(self, size: tuple[int], key: str | None) -> str | None:
        prev_shown = self._eventshown
        self._eventshown = None
        self.clear_event_view()

        if key in self._conf["keybindings"]["new"]:
            self.new(self.focus_date, None)
            key = None

        if self.focus_event:
            if key in self._conf["keybindings"]["delete"]:
                self.toggle_delete()
                key = "down"
            elif key in self._conf["keybindings"]["duplicate"]:
                self.duplicate()
                key = None
            elif key in self._conf["keybindings"]["export"]:
                self.export_event()
                key = None

        rval = super().keypress(size, key)
        if self.focus_event:
            if key in self._conf["keybindings"]["view"] and prev_shown == self.focus_event.recuid:
                # the event in focus is already viewed -> edit
                if self.delete_status(self.focus_event.recuid):
                    self.pane.window.alert(("alert", "This event is marked as deleted"))
                self.edit(self.focus_event.event)
            elif key in self._conf["keybindings"]["external_edit"]:
                self.edit(self.focus_event.event, external_edit=True)
            elif (
                key in self._conf["keybindings"]["view"]
                or self._conf["view"]["event_view_always_visible"]
            ):
                self._eventshown = self.focus_event.recuid
                self.view(self.focus_event.event)
        return rval

    def render(self, a, focus):
        if focus:
            DateListBox.selected_date = None
        return super().render(a, focus)


class EventDisplay(urwid.WidgetWrap):
    """A widget showing one Event()'s details"""

    def __init__(self, conf, event, collection=None) -> None:
        self._conf = conf
        self.collection = collection
        self.event = event
        divider = urwid.Divider(" ")

        lines = []
        lines.append(urwid.Text("Title: " + event.summary))

        # show organizer
        if event.organizer != "":
            lines.append(urwid.Text("Organizer: " + event.organizer))

        if event.location != "":
            lines.append(urwid.Text("Location: " + event.location))

        if event.categories != "":
            lines.append(urwid.Text("Categories: " + event.categories))

        if event.url != "":
            lines.append(urwid.Text("URL: " + event.url))

        if event.attendees != "":
            lines.append(urwid.Text("Attendees:"))
            for attendee in event.attendees.split(", "):
                lines.append(urwid.Text(f"  - {attendee}"))

        # start and end time/date
        if event.allday:
            startstr = event.start_local.strftime(self._conf["locale"]["dateformat"])
            endstr = event.end_local.strftime(self._conf["locale"]["dateformat"])
        else:
            startstr = event.start_local.strftime(
                f"{self._conf['locale']['dateformat']} {self._conf['locale']['timeformat']}"
            )
            if event.start_local.date == event.end_local.date:
                endstr = event.end_local.strftime(self._conf["locale"]["timeformat"])
            else:
                endstr = event.end_local.strftime(
                    f"{self._conf['locale']['dateformat']} {self._conf['locale']['timeformat']}"
                )

        if startstr == endstr:
            lines.append(urwid.Text("Date: " + startstr))
        else:
            lines.append(urwid.Text("Date: " + startstr + " - " + endstr))

        lines.append(urwid.Text("Calendar: " + event.calendar))
        lines.append(divider)

        if event.description != "":
            lines.append(urwid.Text(event.description))

        pile = urwid.Pile(lines)
        urwid.WidgetWrap.__init__(self, urwid.Filler(pile, valign="top"))


class SearchDialog(urwid.WidgetWrap):
    """A Search Dialog Widget"""

    def __init__(self, search_func, abort_func) -> None:
        class Search(Edit):
            def keypress(self, size: tuple[int], key: str | None) -> str | None:
                if key == "enter":
                    search_func(self.text)
                    return None
                else:
                    return super().keypress(size, key)

        search_field = Search("")

        def this_func(_):
            search_func(search_field.text)

        lines = []
        lines.append(urwid.Text("Please enter a search term (Escape cancels):"))
        lines.append(urwid.AttrMap(search_field, "edit", "edit focused"))
        lines.append(urwid.Text(""))
        buttons = NColumns(
            [
                button("Search", on_press=this_func, padding_left=0),
                button("Abort", on_press=abort_func, padding_right=0),
            ]
        )
        lines.append(buttons)
        content = NPile(lines, outermost=True)
        urwid.WidgetWrap.__init__(self, urwid.LineBox(content))


class ClassicView(Pane):
    """default Pane for khal

    showing a CalendarWalker on the left and the eventList + eventviewer/editor
    on the right
    """

    def __init__(self, collection, conf=None, title: str = "", description: str = "") -> None:
        self.init = True
        # Will be set when opening the view inside a Window
        self.window = None
        self._conf = conf
        self.collection = collection
        self._deleted: dict[int, list[str]] = {DeletionType.ALL: [], DeletionType.INSTANCES: []}

        ContainerWidget = linebox[self._conf["view"]["frame"]]
        if self._conf["view"]["dynamic_days"]:
            Walker = DayWalker
        else:
            Walker = StaticDayWalker
        daywalker = Walker(
            dt.date.today(),
            eventcolumn=self,
            conf=self._conf,
            delete_status=self.delete_status,
            collection=self.collection,
        )
        elistbox = DListBox(
            daywalker,
            parent=self,
            conf=self._conf,
            delete_status=self.delete_status,
            toggle_delete_all=self.toggle_delete_all,
            toggle_delete_instance=self.toggle_delete_instance,
            dynamic_days=self._conf["view"]["dynamic_days"],
        )
        self.eventscolumn = ContainerWidget(
            CAttrMap(
                EventColumn(pane=self, elistbox=elistbox),
                "eventcolumn",
                "eventcolumn focus",
            ),
        )
        calendar = CAttrMap(
            CalendarWidget(
                on_date_change=self.eventscolumn.original_widget.set_focus_date,
                keybindings=self._conf["keybindings"],
                on_press=dict.fromkeys(self._conf["keybindings"]["new"], self.new_event),
                firstweekday=self._conf["locale"]["firstweekday"],
                weeknumbers=self._conf["locale"]["weeknumbers"],
                monthdisplay=self._conf["view"]["monthdisplay"],
                get_styles=collection.get_styles,
            ),
            "calendar",
            "calendar focus",
        )
        if self._conf["view"]["dynamic_days"]:
            elistbox.set_focus_date_callback = calendar.set_focus_date
        else:
            elistbox.set_focus_date_callback = lambda _: None
        self.calendar = ContainerWidget(calendar)
        self.lwidth = 31 if self._conf["locale"]["weeknumbers"] == "right" else 28
        if self._conf["view"]["frame"] in ["width", "color"]:
            self.lwidth += 2
        columns = NColumns(
            [(self.lwidth, self.calendar), self.eventscolumn],
            dividechars=0,
            box_columns=[0, 1],
            outermost=True,
        )
        Pane.__init__(self, columns, title=title, description=description)

    def delete_status(self, uid: str) -> DeletionType | None:
        if uid[0] in self._deleted[DeletionType.ALL]:
            return DeletionType.ALL
        elif uid in self._deleted[DeletionType.INSTANCES]:
            return DeletionType.INSTANCES
        else:
            return None

    def toggle_delete_all(self, recuid: tuple[str, str]) -> None:
        uid, _ = recuid
        if uid in self._deleted[DeletionType.ALL]:
            self._deleted[DeletionType.ALL].remove(uid)
        else:
            self._deleted[DeletionType.ALL].append(uid)

    def toggle_delete_instance(self, uid: str) -> None:
        if uid in self._deleted[DeletionType.INSTANCES]:
            self._deleted[DeletionType.INSTANCES].remove(uid)
        else:
            self._deleted[DeletionType.INSTANCES].append(uid)

    def cleanup(self, data):
        """delete all events marked for deletion"""
        # If we delete several recuids from the same vevent, the etag will
        # change after each deletion. As we check for matching etags before
        # deleting, deleting any subsequent recuids will fail.
        # We therefore keep track of the etags of the events we already
        # deleted.
        updated_etags = {}
        for part in self._deleted[DeletionType.ALL]:
            account, href, etag = part.split("\n", 2)
            self.collection.delete(href, etag, account)
        for part, rec_id in self._deleted[DeletionType.INSTANCES]:
            account, href, etag = part.split("\n", 2)
            etag = updated_etags.get(href) or etag
            event = self.collection.delete_instance(href, etag, account, rec_id)
            updated_etags[event.href] = event.etag

    def keypress(self, size: tuple[int], key: str | None) -> str | None:
        binds = self._conf["keybindings"]
        if key in binds["search"]:
            self.search()
        return super().keypress(size, key)

    def search(self):
        """create a search dialog and display it"""
        overlay = urwid.Overlay(
            SearchDialog(self._search, self.window.backtrack),
            self,
            align="center",
            width=("relative", 70),
            valign=("relative", 50),
            height=None,
        )
        self.window.open(overlay)

    def _search(self, search_term: str) -> None:
        """search for events matching `search_term"""
        assert self.window is not None
        self.window.backtrack()
        events = sorted(self.collection.search(search_term))
        event_list = []
        event_list.extend(
            [
                urwid.AttrMap(
                    U_Event(
                        event, relative=False, conf=self._conf, delete_status=self.delete_status
                    ),
                    "calendar " + event.calendar,
                    "reveal focus",
                )
                for event in events
            ]
        )
        events = EventListBox(
            urwid.SimpleFocusListWalker(event_list),
            parent=self.eventscolumn,
            conf=self._conf,
            delete_status=self.delete_status,
            toggle_delete_all=self.toggle_delete_all,
            toggle_delete_instance=self.toggle_delete_instance,
        )
        events = EventColumn(pane=self, elistbox=events)
        ContainerWidget = linebox[self._conf["view"]["frame"]]
        columns = NColumns(
            [(self.lwidth, self.calendar), ContainerWidget(events)],
            dividechars=0,
            box_columns=[0, 0],
            outermost=True,
        )
        pane = Pane(
            columns,
            title=f'Search results for "{search_term}" (Esc for backtrack)',
        )
        pane._conf = self._conf
        columns.set_focus_column(1)
        self.window.open(pane)

    def render(self, size, focus=False):
        rval = super().render(size, focus)
        if self.init:
            # starting with today's events
            self.eventscolumn.current_date = dt.date.today()
            self.init = False
        return rval

    def new_event(self, date, end):
        """create a new event starting on date and ending on end (if given)"""
        self.eventscolumn.original_widget.new(date, end)


def _urwid_palette_entry(
    name: str,
    color: str,
    hmethod: str,
    color_mode: Literal["256colors", "rgb"],
    foreground: str = "",
    background: str = "",
) -> tuple[str, str, str, str, str, str]:
    """Create an urwid compatible palette entry.

    :param name: name of the new attribute in the palette
    :param color: color for the new attribute
    :param hmethod: which highlighting mode to use, foreground or background
    :param color_mode: which color mode we are in, if we are in 256-color mode,
    we transform 24-bit/RGB colors to a (somewhat) matching 256-color set color
    :param foreground: the foreground color to apply if we use background highlighting method
    :param background: the background color to apply if we use foreground highlighting method

    :returns: an urwid palette entry
    """
    from khal.terminal import COLORS

    if color == "" or color in COLORS or color is None:
        # Named colors already use urwid names, no need to change anything.
        pass
    elif color.isdigit():
        # Colors from the 256 color palette need to be prefixed with h in
        # urwid.
        color = "h" + color
    elif color_mode == "256color":
        # Convert to some color on the 256 color palette that might resemble
        # the 24-bit color.
        # First, generate the palette (indices 16-255 only). This assumes, that
        # the terminal actually uses the same palette, which may or may not be
        # the case.
        colors = {}
        # Colorcube
        colorlevels = (0x00, 0x5F, 0x87, 0xAF, 0xD7, 0xFF)
        for r in range(6):
            for g in range(6):
                for b in range(6):
                    colors[r * 36 + g * 6 + b + 16] = (
                        colorlevels[r],
                        colorlevels[g],
                        colorlevels[b],
                    )
        # Grayscale
        graylevels = [0x08 + 10 * i for i in range(24)]
        for c in range(24):
            colors[232 + c] = (graylevels[c],) * 3
        # Parse the HTML-style color into the variables r, g, b.
        if len(color) == 4:
            # e.g. #ABC, equivalent to #AABBCC
            r = int(color[1] * 2, 16)
            g = int(color[2] * 2, 16)
            b = int(color[3] * 2, 16)
        else:
            # e.g. #AABBCC
            r = int(color[1:3], 16)
            g = int(color[3:5], 16)
            b = int(color[5:7], 16)
        # Now, find the color with the least distance to the requested color.
        best = None
        bestdist = 0.0
        for index, rgb in colors.items():
            # This is the euclidean distance metric. It is quick, simple and
            # wrong (in the sense of human color perception). However, any
            # serious color distance metric would be way more complicated.
            dist = (r - rgb[0]) ** 2 + (g - rgb[1]) ** 2 + (b - rgb[2]) ** 2
            if best is None or dist < bestdist:
                best = index
                bestdist = dist
        color = "h" + str(best)
    # We unconditionally add the color to the high color slot. It seems to work
    # in lower color terminals as well.
    if hmethod in ["fg", "foreground"]:
        return (name, "", "", "", color, background)
    else:
        return (name, "", "", "", foreground, color)


def _add_calendar_colors(
    palette: list[tuple[str, ...]],
    collection: "CalendarCollection",
    color_mode: Literal["256colors", "rgb"],
    base: str | None = None,
    attr_template: str = "calendar {}",
) -> list[tuple[str, ...]]:
    """Add the colors for the defined calendars to the palette.

    We support setting a fixed background or foreground color that we extract
    from a giving attribute

    :param palette: the base palette
    :param collection:
    :param color_mode: which color mode we are in
    :param base: the attribute to extract the background and foreground color from
    :param attr_template: the template to use for the attribute name
    :returns: the modified palette
    """
    bg_color, fg_color = "", ""
    for attr in palette:
        if base and attr[0] == base:
            if color_mode == "rgb" and len(attr) >= 5:
                bg_color = attr[5]
                fg_color = attr[4]
            else:
                bg_color = attr[2]
                fg_color = attr[1]

    for cal in collection.calendars:
        if cal["color"] == "":
            # No color set for this calendar, use default_color instead.
            color = collection.default_color
        else:
            color = cal["color"]

        # In case the color contains an alpha value, remove it for urwid.
        # eg '#RRGGBBAA' -> '#RRGGBB' and '#RGBA' -> '#RGB'.
        if color and len(color) == 9 and color[0] == "#":
            color = color[0:7]
        elif color and len(color) == 5 and color[0] == "#":
            color = color[0:4]

        entry = _urwid_palette_entry(
            attr_template.format(cal["name"]),
            color,
            collection.hmethod,
            color_mode=color_mode,
            foreground=fg_color,
            background=bg_color,
        )
        palette.append(entry)

    entry = _urwid_palette_entry(
        "highlight_days_color",
        collection.color,
        collection.hmethod,
        color_mode=color_mode,
        foreground=fg_color,
        background=bg_color,
    )
    palette.append(entry)
    entry = _urwid_palette_entry(
        "highlight_days_multiple",
        collection.multiple,
        collection.hmethod,
        color_mode=color_mode,
        foreground=fg_color,
        background=bg_color,
    )
    palette.append(entry)

    return palette


def start_pane(
    pane,
    callback,
    program_info="",
    quit_keys=None,
    color_mode: Literal["rgb", "256colors"] = "rgb",
):
    """Open the user interface with the given initial pane."""
    # We don't validate the themes in settings.spec but instead here
    # first try to load built-in themes, then try to load themes from
    # plugins
    theme = colors.themes.get(pane._conf["view"]["theme"])
    if theme is None:
        theme = plugins.THEMES.get(pane._conf["view"]["theme"])
    if theme is None:
        logger.fatal(f"Invalid theme {pane._conf['view']['theme']} configured")
        logger.fatal(f"Available themes are: {', '.join(colors.themes.keys())}")
        raise FatalError

    quit_keys = quit_keys or ["q"]

    frame = Window(
        footer=program_info + f" | {quit_keys[0]}: quit, ?: help",
        quit_keys=quit_keys,
    )

    class LogPaneFormatter(logging.Formatter):
        def get_prefix(self, level):
            if level >= 50:
                return "CRITICAL"
            if level >= 40:
                return "ERROR"
            if level >= 30:
                return "WARNING"
            if level >= 20:
                return "INFO"
            else:
                return "DEBUG"

        def format(self, record) -> str:
            return f"{self.get_prefix(record.levelno)}: {record.msg}"

    class HeaderFormatter(LogPaneFormatter):
        def format(self, record):
            return (
                super().format(record)[:30] + "... "
                f"[Press `{pane._conf['keybindings']['log'][0]}` to view log]"
            )

    class LogPaneHandler(logging.Handler):
        def emit(self, record):
            frame.log(self.format(record))

    class LogHeaderHandler(logging.Handler):
        def emit(self, record):
            frame.alert(self.format(record))

    if len(logger.handlers) > 0 and not isinstance(logger.handlers[-1], logging.FileHandler):
        logger.handlers.pop()

    pane_handler = LogPaneHandler()
    pane_handler.setFormatter(LogPaneFormatter())
    logger.addHandler(pane_handler)

    header_handler = LogHeaderHandler()
    header_handler.setFormatter(HeaderFormatter())
    logger.addHandler(header_handler)

    frame.open(pane, callback)
    palette = _add_calendar_colors(
        theme,
        pane.collection,
        color_mode=color_mode,
        base="calendar",
        attr_template="calendar {}",
    )
    palette = _add_calendar_colors(
        palette,
        pane.collection,
        color_mode=color_mode,
        base="popupbg",
        attr_template="calendar {} popup",
    )

    def merge_palettes(pallete_a, pallete_b) -> list[tuple[str, ...]]:
        """Merge two palettes together, with the second palette taking priority."""
        merged = {}
        for entry in pallete_a:
            merged[entry[0]] = entry
        for entry in pallete_b:
            merged[entry[0]] = entry
        return list(merged.values())

    overwrite = [(key, *values) for key, values in pane._conf["palette"].items()]
    palette = merge_palettes(palette, overwrite)
    loop = urwid.MainLoop(
        widget=frame,
        palette=palette,
        unhandled_input=frame.on_key_press,
        pop_ups=True,
        handle_mouse=pane._conf["default"]["enable_mouse"],
    )
    frame.loop = loop

    def redraw_today(loop, pane, meta=None):
        meta = meta or {"last_today": None}
        # XXX TODO this currently assumes, today moves forward by exactly one
        # day, but it could either move forward more (suspend-to-disk/ram) or
        # even move backwards
        today = dt.date.today()
        if meta["last_today"] != today:
            meta["last_today"] = today
            pane.calendar.original_widget.reset_styles_range(today - dt.timedelta(days=1), today)
            pane.eventscolumn.original_widget.update_date_line()
        loop.set_alarm_in(60, redraw_today, pane)

    loop.set_alarm_in(60, redraw_today, pane)

    def check_for_updates(loop, pane):
        if pane.collection.needs_update():
            pane.window.alert("detected external vdir modification, updating...")
            pane.collection.update_db()
            pane.eventscolumn.base_widget.update(None, None, everything=True)
            pane.window.alert("detected external vdir modification, updated.")
        loop.set_alarm_in(60, check_for_updates, pane)

    loop.set_alarm_in(60, check_for_updates, pane)

    colors_ = 2**24 if color_mode == "rgb" else 256
    loop.screen.set_terminal_properties(
        colors=colors_,
        bright_is_bold=pane._conf["view"]["bold_for_light_color"],
    )

    def ctrl_c(signum, f):
        raise urwid.ExitMainLoop()

    signal.signal(signal.SIGINT, ctrl_c)
    try:
        loop.run()
    except Exception:
        import traceback

        tb = traceback.format_exc()
        try:  # Try to leave terminal in usable state
            loop.stop()
        except Exception:
            pass
        print(tb)
        sys.exit(1)



# Copyright (c) 2013-2022 khal contributors
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


class Error(Exception):
    """base class for all of khal's Exceptions"""


class FatalError(Error):
    """execution cannot continue"""


class DateTimeParseError(FatalError):
    pass


class ConfigurationError(FatalError):
    pass


class UnsupportedFeatureError(Error):
    """something Failed but we know why"""


class UnsupportedRecurrence(Error):
    """raised if the RRULE is not understood by dateutil.rrule"""


class InvalidDate(Error):
    pass
